/* Auto-generated by generate-wrappers.py script. Do not modify */

#if defined(__arm__) || defined(__aarch64__)
#include <q8dwconv/up8x9-neon-per-channel.c>
#endif /* defined(__arm__) || defined(__aarch64__) */
