/* styles needed for the Google Search button */

.gsc-overflow-hidden {
    overflow: visible !important;
}

.gsc-control-cse {
   padding-left: 0px !important;
   padding-bottom: 0px !important;
}

.gsc-search-button .gsc-search-button-v2:focus {
   border: transparent !important;
   outline: none;
   box-shadow: none;
}

.gsc-search-button-v2:active {
   border: none !important;
}

.gsc-search-button-v2 {
   border: none !important;
}

/* End of Google Search button styles */

/* Classes needed for the survey link*/

.pytorch-left-menu-search { /* This is needed so the distance between the search and menu is not too big */
  margin-bottom: 0px;
}

.survey-link {
  color: #262626;
}

.pytorch-left-menu-search .survey-link a {
  color: #262626;
  text-decoration: underline;
  font-weight: 500;
}

/* End of classes needed for the survey banner*/


/* search radio button*/
input[type="radio"] {
    accent-color: #ee4c2c;
}
.gsst_b {
    display: none;
}
#gsc-i-id1 {
    height: 1.5rem;
    text-indent: 12px !important;
    font-size: 1rem !important;
    font-family: "FreightSansi";
    background-image: url(../images/search-icon.svg) !important;
    background-repeat: no-repeat !important;
    background-size: 18px 18px !important;
    background-position: 5px 0px !important;
    padding-left: 20px !important;
}
#gsc-i-id1::placeholder {
  font-family: 'FreightSans';
  font-size: 1rem;
  color: #262626;
}
.gsc-control-cse {
   padding: 0 !important;
   border-radius: 0px !important;
   border: none !important;;
   overflow: hidden;
}
#___gcse_0 {
    height: 44px !important;
    padding: 0 !important;
}

@media (max-width: 768px) {
  #___gcse_0 {
    height: 34px !important;
  }
}

table.gsc-search-box td.gsc-input {
    padding-right: 0 !important;
}
table.gsc-search-box td {
    height: 44px;
    margin-bottom: 0 !important;
    padding-bottom: 0 !important;
}

@media (max-width: 768px) {
  table.gsc-search-box td {
    height: 34px;
  }
}

.gsc-search-button-v2 {
    display: none;
}
.gs_id50 {
    width: 308px;
}

.gsib_a {
    padding: 0px 8px 4px 9px !important;
}
.gsc-input-box {
    border-radius: 0px !important;
    border: none !important;
}

@media (max-width: 768px) {
  .gsc-input-box {
     padding-top: 0px;
     padding-bottom: 0px;
  }
}

@media (max-width: 768px) {
   .searchbox {
     height: 34px;
}


form.gsc-search-box {
    margin-bottom 0px;
}

.pytorch-left-menu-search input[type=text] {
    background-image: url("../images/search-icon.svg");
}

.gsc-overflow-hidden {
  overflow: visible !important;
}

.form#searchForm {
  padding-bottom: 0;
  margin-bottom: 0;
}
